// 8-LED 4-cell nickel battery tester for the LGT8F micro board
// ceptimus  2022-08-13
// 1k2 2k2 divider converts 5.36V (1.34V * 4) to 1.8918V and 4.8V (1.2V * 4) to 1.6941V
// suits 2V048 internal reference range

// a flat Nickel cell is about 1.2 volts or below.  7 steps of 0.02 volts per cell takes us up to 1.34V for fully charged.  For four cells we then have 4800mV with 80mV steps 
// maximum voltage for a flat battery
#define FLAT 4800
// millivolt increase for each subsequent LED
#define STEP 80
#define FULL (FLAT + 7 * STEP)

#include <EEPROM.h>

// remove commenting-out of DEBUG to enable Serial printing
//#define DEBUG

// const int leds[] = { 8, 10, 12, A0, A1, A4, 3, 2 }; // low to high volts only one lit at once (LOW) with shared resistor (220R) (use this one if LED colours in wrong order)
const int leds[] = { 2, 3, A4, A1, A0, 12, 10, 8 }; // low to high volts only one lit at once (LOW) with shared resistor (220R)

void allLEDsOff() {
  for (uint8_t i = 0; i < 8; i++) {
    digitalWrite(leds[i], HIGH);
    pinMode(leds[i], OUTPUT);
  }
}

uint16_t flatmv,  fullmv;

void readCalibrationValuesFromEeprom() { // with sanity check and defaults for insane values
  flatmv = EEPROM.read(0) + EEPROM.read(1) * 256;
  fullmv = EEPROM.read(2) + EEPROM.read(3) * 256;
  if (flatmv >= fullmv || flatmv < 2048 || fullmv > 4000) { // default to theoretical values with 1.2k and 2.2k potential divider resistors
    flatmv = 3388; // 4.80 * 1200 / (1200 + 2200) / 2.048 * 4096
    fullmv = 3784; // 5.36 * 1200 / (1200 + 2200) / 2.048 * 4096
  }
}

void setup() {
#ifdef DEBUG
  Serial.begin(115200);
#endif
  analogReference(INTERNAL2V048);
  pinMode(7, INPUT_PULLUP); // briefly link  7 to GND to calibrate  low (flat) voltage.  Recommend doing this at 4.800 V
  pinMode(13, INPUT_PULLUP);// briefly link 13 to GND to calibrate high (full) voltage.  Recommend doing this at 5.360 V
  readCalibrationValuesFromEeprom();
#ifdef DEBUG
  Serial.print("flatmv = ");
  Serial.println(flatmv);
  Serial.print("fullmv = ");
  Serial.println(fullmv);
  delay(2000);
#endif
}

void eepromUpdate(uint16_t addr, uint8_t value) { // LGT8F has no native update (August 2022)
  uint8_t existing = EEPROM.read(addr);
  if (value != existing) {
    EEPROM.write(addr, value);
  }
}

void updateEeprom() { // write current flatmv and fullmv to first four bytes of EEPROM
  eepromUpdate(0, flatmv & 0x00FF);
  eepromUpdate(1, flatmv >> 8);
  eepromUpdate(2, fullmv & 0x00FF);
  eepromUpdate(3, fullmv >> 8);
  // display a pattern on the LEDS to 1, show something happened, and 2, provide a delay before updating EEPROM again if they're still shorting the 'calibrate' wire
  for (uint8_t i = 0; i < 8; i++) {
    allLEDsOff();
    delay(100);
    digitalWrite(leds[i], LOW);
    delay(200);
  }
  readCalibrationValuesFromEeprom();
}

void loop() {
  uint16_t a = analogRead(A5);
  uint16_t mv = map(a, flatmv, fullmv, FLAT, FULL);
#ifdef DEBUG
  Serial.print(a);
  Serial.print("   ");
  Serial.println(mv);
#endif
  allLEDsOff();
  if      (mv <= FLAT + 0 * STEP) digitalWrite(leds[0], LOW);
  else if (mv <= FLAT + 1 * STEP) digitalWrite(leds[1], LOW);
  else if (mv <= FLAT + 2 * STEP) digitalWrite(leds[2], LOW);
  else if (mv <= FLAT + 3 * STEP) digitalWrite(leds[3], LOW);
  else if (mv <= FLAT + 4 * STEP) digitalWrite(leds[4], LOW);
  else if (mv <= FLAT + 5 * STEP) digitalWrite(leds[5], LOW);
  else if (mv <= FLAT + 6 * STEP) digitalWrite(leds[6], LOW);
  else                            digitalWrite(leds[7], LOW);

  if (!digitalRead(7)) { // calibrate flatmv
    flatmv = a;
    updateEeprom();
  }

  if (!digitalRead(13)) { // calibrate fullmv
    fullmv = a;
    updateEeprom();
  }
  
  delay(50);
}
